@@ -1,7 +1,8 @@
 ;;; fedi-post.el --- Minor mode for posting to fediverse services  -*- lexical-binding: t -*-
 
 ;; Copyright (C) 2020-2023 Marty Hiatt
-;; Author: Marty Hiatt <martianhiatus@riseup.net> and mastodon.el authors
+;; Author: Marty Hiatt <mousebot@disroot.org> and mastodon.el authors
+;; Version: 1.0.0
 ;; Homepage: https://codeberg.org/martianh/fedi.el
 
 ;; This file is not part of GNU Emacs.
@@ -118,29 +119,38 @@ face. no-label is optional.")
         (+ ; breaks groups with instance handles!
          ?@
          (group-n 2 ; = username only
-           (* (any ?- ?_ ?. "A-Z" "a-z" "0-9" )))
+           (* (any ?- ?_ ;?. ;; this . breaks word-boundary at end
+                   "A-Z" "a-z" "0-9")))
          (? ?@
             (group-n 3 ; = optional domain
               (* (not (any "\n" "\t" " ")))))))
+      ;; FIXME: fails with full stop? but we must exclude it! other regexes
+      ;; don't fail with full stops!
       (| "'" word-boundary)) ; boundary or possessive
-  "Regex for a handle, e.g. @user@instance.com.
+  "Regex for a handle, e.g. @user or @user@instance.com.
 Group 1 is for completion at point functions. Group 2 and 3 are
 for forming a URL.")
 
 (defvar fedi-post-tag-regex
   (rx (| (any ?\( "\n" "\t" " ") bol)
-      (group-n 1 ?# (+ (any "A-Z" "a-z" "0-9")))
+      (group-n 1 ?#
+               (group-n 2
+                 (one-or-more (any "A-Z" "a-z" "0-9"))))
       (| "'" word-boundary)) ; boundary or possessive
   "Regex for a tag. Group 1 is for completion at point functions.")
 
 (defvar fedi-post-url-regex
   ;; adapted from `ffap-url-regexp'
   (concat
-   "\\(?1:\\(news\\(post\\)?:\\|mailto:\\|file:\\|\\(ftp\\|https?\\|telnet\\|gopher\\|www\\|wais\\)://\\)"
-                                        ; uri prefix
-   "[^ \n\t]*\\)" ; any old thing, that is, i.e. we allow invalid/unwise chars
-   "\\b"))
-                                        ; boundary
+   "\\(?1:\\(news\\(post\\)?:\\|mailto:\\|file:\\|\\(ftp\\|https?\\|telnet\\|gopher\\|www\\|wais\\)://\\)" ;; uri prefix
+   "[^ \n\t,]*\\)" ;; any old thing, that is, i.e. we allow invalid/unwise chars
+   "\\(\\b\\|\\.\\)")) ;; boundary or terminating period
+
+(defvar fedi-post-commit-regex
+  (rx (| (any ?\( "\n" "\t" " ") bol)
+      (group-n 1 (>= 7 hex-digit)) ;; 7 or more hex
+      (| "'" word-boundary))
+  "Regex for a commit ref, a 7-digit hex value.")
 
 
 ;;; MODE MAP
@@ -219,11 +229,17 @@ Buffer-local variable `fedi-post-previous-window-config' holds the config."
               (match-end 1))))))
 
 (defun fedi-post--return-capf (regex completion-fun &optional
-                                     annot-fun _affix-fun exit-fun)
+                                     annot-fun _affix-fun exit-fun
+                                     category)
   "Return a completion at point function.
 REGEX is used to get the item before point.
+
 COMPLETION-FUN takes two args, start and end bounds of item
-before point, and returns a completion table.
+before point, and returns a completion table. Nota bene that for
+completion to work correctly, COMPLETION-FUN must return data
+containing candidates that will literally match the preceding
+search string, including any prefixes like @ or #.
+
 ANNOT-FUN takes one arg, a candidate, and returns an annotation
 for it.
 AFFIX-FUN is currently unused, it would be :affixation-function.
@@ -249,12 +265,15 @@ and a status."
             ;; (funcall affix-fun cands))
             ;; FIXME: we "should" use :affixation-function for this but i
             ;; can't get it to work so use an exit-fun hack:
+            :category category
             :exit-function
-            (lambda (str status)
-              (funcall exit-fun str status))
+            (when exit-fun
+              (lambda (str status)
+                (funcall exit-fun str status)))
             :annotation-function
-            (lambda (cand)
-              (concat " " (funcall annot-fun cand)))))))
+            (when annot-fun
+              (lambda (cand)
+                (concat " " (funcall annot-fun cand))))))))
 
 (defun fedi-post--mentions-annotation-fun (candidate)
   "Given a handle completion CANDIDATE, return its annotation string, a username."
@@ -380,7 +399,7 @@ LONGEST is the length of the longest binding."
   (propertize str
               'face 'fedi-post-docs-face))
 
-(defun fedi-post--make-mode-docs (&optional mode prefix type)
+(defun fedi-post--make-mode-docs (&optional mode prefix type edit)
   "Create formatted documentation text for MODE or fedi-post-mode.
 PREFIX is a string corresponding to the prefix of the minor mode
 enabled. It is used for constructing clean keybinding
@@ -394,7 +413,8 @@ descriptions."
     (concat
      (fedi-post-comment
       (format
-       " Compose a new %s here. The following keybindings are available:" type))
+       " %s %s here. The following keybindings are available:"
+       (if edit "Edit your" "Compose a new") type))
      (mapconcat #'identity
                 (fedi-post--formatted-kbinds-pairs
                  (fedi-post--format-kbinds kbinds prefix)
@@ -402,16 +422,17 @@ descriptions."
                 nil))))
 
 (defun fedi-post--concat-fields (fields)
-  "Concat FIELDS for compose buffer docs."
+  "Concat FIELDS for compose buffer docs.
+The property added here is used by `fedi-post--update-status-fields'
+to update status fields."
   (cl-loop for item in fields
-           for field = (alist-get 'name item)
-           concat (propertize (capitalize field)
-                              (intern
-                               (concat "post-" (downcase field)))
-                              t)))
+           for name = (alist-get 'name item)
+           for prop = (alist-get 'prop item)
+           concat (propertize (capitalize name)
+                              (if (consp prop) (car prop) prop) t)))
 
 (defun fedi-post--display-docs-and-status-fields (&optional mode prefix
-                                                            fields type)
+                                                            fields type edit)
   "Insert propertized text with documentation about MODE or `fedi-post-mode'.
 Also includes and the status fields which will get updated based
 on the status of NSFW, language, media attachments, etc.
@@ -423,7 +444,7 @@ FIELDS is a list of alists of fields to add, using `fedi-post--concat-fields'."
          "|=================================================================|"))
     (insert
      (concat
-      (fedi-post--make-mode-docs mode prefix type) "\n"
+      (fedi-post--make-mode-docs mode prefix type edit) "\n"
       (fedi-post-comment divider) "\n"
       (propertize
        (concat
@@ -459,24 +480,36 @@ FIELDS is a list of alists of fields to add, using `fedi-post--concat-fields'."
     (length (buffer-substring (point-min) (point-max)))))
 
 (defun fedi-post--update-status-field (item)
-  "ITEM."
+  "Update status field for ITEM.
+ITEM is an alist with name, prop, item-var, no-label and face fields.
+Item-var, containing the data to be displayed, can be a string, or a
+cons cell, or a list. If the latter two, the first element is displayed."
   (let-alist item
-    (let ((region (fedi--find-property-range .prop (point-min))))
-      (add-text-properties (car region) (cdr region)
-                           (list 'display
-                                 (if (eval .item-var)
-                                     (format
-                                      (concat
-                                       (if .no-label
-                                           ""
-                                         (concat (capitalize .name)
-                                                 ": "))
-                                       (propertize "%s"
-                                                   'face .face)
-                                       " ⋅ ")
-                                      (eval .item-var))
-                                   "")
-                                 'face 'fedi-post-docs-face)))))
+    (let ((region (fedi--find-property-range .prop (point-min)))
+          (val (symbol-value .item-var)))
+      (add-text-properties
+       (car region) (cdr region)
+       (list 'display
+             (if (not val)
+                 ""
+               (format
+                (concat
+                 (if .no-label
+                     ""
+                   (concat (capitalize .name)
+                           ": "))
+                 (propertize "%s"
+                             'face .face)
+                 " ⋅ ")
+                (cond ((proper-list-p val) ;; list
+                       (mapconcat (lambda (x)
+                                    ;; for alists, concat the cars:
+                                    (if (consp x) (car x) x))
+                                  val " "))
+                      ((consp val) ;; cons, use car
+                       (car val))
+                      (t val))))
+             'face 'fedi-post-docs-face)))))
 
 (defun fedi-post--update-status-fields (&rest _args)
   "Update the status fields in the header based on the current state."
@@ -548,8 +581,14 @@ Added to `after-change-functions'."
   "Return t if compose buffer is current."
   (let ((buf (buffer-name (current-buffer))))
     ;; TODO: generalize:
+    ;; (let ((new-or-edit '("new" . "edit"))
+    ;; (types '("post" "issue" "comment")))
     (or (equal "*new post*" buf)
-        (equal "*edit post*" buf))))
+        (equal "*edit post*" buf)
+        (equal "*new issue*" buf)
+        (equal "*edit issue*" buf)
+        (equal "*new comment*" buf)
+        (equal "*edit comment*" buf))))
 
 (defun fedi-post--fill-reply-in-compose ()
   "Fill reply text in compose buffer to the width of the divider."
@@ -557,16 +596,32 @@ Added to `after-change-functions'."
     (save-match-data
       (let* ((fill-column 67))
         (goto-char (point-min))
-        (when-let ((prop (text-property-search-forward 'post-reply)))
+        (when-let* ((prop (text-property-search-forward 'post-reply)))
           (fill-region (prop-match-beginning prop)
                        (point)))))))
 
+(defun fedi-post--render-reply-region-str (str)
+  "Refill STR and prefix all lines with >, as reply-quote text."
+  (with-temp-buffer
+    (insert str)
+    ;; unfill first:
+    (let ((fill-column (point-max)))
+      (fill-region (point-min) (point-max)))
+    ;; then fill:
+    (fill-region (point-min) (point-max))
+    ;; add our own prefix, pauschal:
+    (goto-char (point-min))
+    (save-match-data
+      (while (re-search-forward "^" nil t)
+        (replace-match " > ")))
+    (buffer-substring-no-properties (point-min) (point-max))))
 
 
 ;;; COMPOSE BUFFER FUNCTION
 
 (defun fedi-post--compose-buffer
-    (&optional edit major minor prefix type capf-funs fields)
+    (&optional edit major minor prefix type capf-funs fields
+               init-text reply-text buf-prefix autocomplete)
   "Create a new buffer to capture text for a new post.
 EDIT means we are editing an existing post, not composing a new one.
 MAJOR is the major mode to enable.
@@ -579,10 +634,11 @@ TYPE is a string for the buffer name.
 FIELDS is a list of alists containing status fields for bindings
 and options display. Each alist should have a name, prop,
 item-var and face elements. Element name should be a hyphen-separated
-string, the other elements should be symbols."
+string, the other elements should be symbols.
+BUF-PREFIX is a string to prepend to the buffer name."
   (let* ((buffer-name (if edit
-                          (format "*edit %s*" type)
-                        (format "*new %s*" type)))
+                          (format "*%sedit %s*" buf-prefix type)
+                        (format "*%snew %s*" buf-prefix type)))
          (buffer-exists (get-buffer buffer-name))
          (buffer (or buffer-exists (get-buffer-create buffer-name)))
          (inhibit-read-only t)
@@ -604,7 +660,7 @@ string, the other elements should be symbols."
         ;; (setq markdown-mode-hook (delete 'variable-pitch-mode markdown-mode-hook))
         (variable-pitch-mode -1)))
     (unless buffer-exists
-      (fedi-post--display-docs-and-status-fields minor prefix fields type))
+      (fedi-post--display-docs-and-status-fields minor prefix fields type edit))
     ;; set up completion:
     (when fedi-post--enable-completion
       (set (make-local-variable 'completion-at-point-functions)
@@ -617,7 +673,11 @@ string, the other elements should be symbols."
         (declare-function company-mode-on "company")
         (set (make-local-variable 'company-backends)
              (add-to-list 'company-backends 'company-capf))
-        (company-mode-on)))
+        (company-mode-on))
+      ;; corfu
+      (when (require 'corfu nil :no-error)
+        (when autocomplete (setq-local corfu-auto t))
+        (corfu--on)))
     ;; after-change:
     (make-local-variable 'after-change-functions)
     ;; (cl-pushnew #'fedi-post--save-post-text after-change-functions)
@@ -638,7 +698,18 @@ string, the other elements should be symbols."
     ;; FIXME: this is incompat with propertize-tags-and-handles
     ;; we would need to add our own propertizing to md-mode font-locking
     (when (eq major 'markdown-mode)
-      (cl-pushnew #'fedi-post-fontify-body-region after-change-functions))))
+      (cl-pushnew #'fedi-post-fontify-body-region after-change-functions))
+    ;; dir locals
+    (let ((enable-local-variables :all))
+      (hack-dir-local-variables-non-file-buffer))
+    ;; init and reply text
+    (when init-text
+      (insert init-text)
+      (delete-trailing-whitespace))
+    (when reply-text
+      (insert "\n"
+              (fedi-post--render-reply-region-str reply-text)
+              "\n"))))
 
 (defun fedi-post-fontify-body-region (&rest _args)
   "Call `font-lock-fontify-region' on post body.
