@@ -1,9 +1,8 @@
 ;;; fedi.el --- Helper functions for fediverse clients  -*- lexical-binding: t -*-
 
 ;; Copyright (C) 2020-2023 Marty Hiatt and mastodon.el authors
-;; Author: Marty Hiatt <martianhiatus@riseup.net>
-;; Package-Version: 0.2
-;; Package-Revision: ffcb84bb132a
+;; Author: Marty Hiatt <mousebot@disroot.org>
+;; Version: 0.2
 ;; Package-Requires: ((emacs "28.1") (markdown-mode "2.5"))
 ;; Homepage: https://codeberg.org/martianh/fedi.el
 
@@ -25,7 +24,7 @@
 ;;; Commentary:
 
 ;; fedi.el provides a bunch of untility functions and macros (originally
-;; adopted from popular mastodon.el
+;; adopted from the popular mastodon.el
 ;; <https://codeberg.org/martianh/mastodon.el> package) to make it easy to
 ;; write clients and interfaces for for JSON APIs.
 
@@ -127,7 +126,8 @@ Used to construct function names in `fedi-request'.")
 ;; Still, it solves the problem of the server rejecting nil param values.
 (defun fedi-arg-when-expr (arg &optional coerce)
   "Return a cons of a string and a symbol type of ARG.
-Also replace _ with - (for Lemmy's type_ param)."
+Also replace _ with - (for Lemmy's type_ param).
+If COERCE, make numbers strings."
   (let ((str
          (string-replace "-" "_" ; for "type_" etc.
                          (symbol-name arg))))
@@ -157,6 +157,57 @@ Also replace _ with - (for Lemmy's type_ param)."
 ;;                          (symbol-name arg))))
 ;;     `(when ,arg (cons ,str "true"))))
 
+(defmacro fedi-opt-params (&rest params)
+  "From PARAMS, a list of symbols, create an alist of parameters.
+Used to conditionally create fields in the parameters alist.
+
+A param can also be an expression, in which case the car should be the
+symbol name of the param as used locally. The cdr should be a plist
+that may contain the fields :boolean, :alias, :when and :list.
+:boolean should be a string, either \"true\" or \"false\".
+:alias should be the name of the parameter as it is on the server.
+:when should be a condition clause to test against rather than the mere
+value of the parameter symbol
+:list should be a list of values for the parameter.
+
+For example:
+
+\(fedi-opt-params (query :alias \"q\") (topic :boolean \"true\")
+                  uid (mode :when (member mode fj-search-modes))
+                  (include-desc :alias \"includeDesc\"
+                                :boolean \"true\")
+                  (list :list (\"123\" \"abc\")
+                  order page limit).
+
+To use this in combination with mandatory parameters, you can do this:
+
+\(append `((\"mandatory\" . ,value))
+          (fedi-opt-params a b c d))."
+  (declare (debug t))
+  `(append ,@(fedi--opt-params-whens params)))
+
+(defun fedi--opt-params-whens (params)
+  "Return a when clause for each of PARAMS, a list of symbols."
+  (cl-loop for x in params
+           collect (fedi--opt-param-expr x)))
+
+(defun fedi--opt-param-expr (param)
+  "For PARAM, return a when expression.
+It takes the form:
+\(when param \\='(\"param\" . param)).
+Param itself can also be an expression. See `fedi-opt-params' for
+details."
+  (if (consp param)
+      (let* ((name (car param))
+             (boolean (plist-get (cdr param) :boolean))
+             (alias (plist-get (cdr param) :alias))
+             (clause (plist-get (cdr param) :when))
+             (list (plist-get (cdr param) :list))
+             (value (or list boolean name))
+             (str (or alias (symbol-name name))))
+        `(when ,(or clause name)
+           `((,,str . ,,value))))
+    `(when ,param `((,(symbol-name ',param) . ,,param)))))
 
 ;;; BUFFER MACRO
 
@@ -170,7 +221,11 @@ than `switch-to-buffer'."
   `(with-current-buffer (get-buffer-create ,buffer)
      (let ((inhibit-read-only t))
        (erase-buffer)
-       (funcall ,mode-fun)
+       (unless (equal major-mode ,mode-fun)
+         (funcall ,mode-fun))
+       ;; FIXME: make this conditional: if prev buffer is same as new one,
+       ;; then no other window, if different, then other window. that makes it
+       ;; so that if we are reloading the same view, we don't other-window
        (if ,other-window
            (switch-to-buffer-other-window ,buffer)
          ;; (switch-to-buffer ,buffer))
@@ -181,7 +236,8 @@ than `switch-to-buffer'."
 ;;; NAV
 
 (defun fedi--goto-pos (fun prop &optional refresh pos)
-  "Search for item with FUN.
+  "Search for item, moving with FUN.
+If PROP is not found after moving, recur.
 If search returns nil, execute REFRESH function.
 Optionally start from POS."
   (let* ((npos (funcall fun
@@ -209,10 +265,13 @@ Optionally start from POS."
 (defun fedi-next-tab-item (&optional previous prop)
   "Move to the next interesting item.
 This could be the next toot, link, or image; whichever comes first.
-Don't move if nothing else to move to is found, i.e. near the end of the buffer.
+Don't move if nothing else to move to is found, i.e. near the end of the
+buffer.
 This also skips tab items in invisible text, i.e. hidden spoiler text.
 PREVIOUS means move to previous item.
-PROP is the text property to search for."
+PROP is the text property to search for.
+Returns nil if nothing found, returns the value of point moved to if
+something found."
   (interactive)
   (let (next-range
         (search-pos (point)))
@@ -226,10 +285,12 @@ PROP is the text property to search for."
       ;; do nothing, all the action is in the while condition
       )
     (if (null next-range)
-        (message "Nothing else here.")
-      (goto-char (car next-range))
-      (if-let ((hecho (fedi--property 'help-echo)))
-          (message "%s" hecho)))))
+        (prog1 nil ;; return nil if nothing (so we can use in or clause)
+          (message "Nothing else here."))
+      (prog1 ;; return point not nil if we moved:
+          (goto-char (car next-range))
+        (if-let* ((hecho (fedi--property 'help-echo)))
+            (message "%s" hecho))))))
 
 (defun fedi-previous-tab-item ()
   "Move to the previous interesting item.
@@ -243,9 +304,9 @@ text, i.e. hidden spoiler text."
 ;;; HEADINGS
 
 (defvar fedi-horiz-bar
-  (if (char-displayable-p ?―)
-      (make-string 12 ?―)
-    (make-string 12 ?-)))
+  (let ((count 16))
+    (make-string count
+                 (if (char-displayable-p ?―) ?― ?-))))
 
 (defun fedi-format-heading (name)
   "Format a heading for NAME, a string."
@@ -254,9 +315,9 @@ text, i.e. hidden spoiler text."
                  name))
          (name (string-replace "-" " " name)))
     (propertize
-     (concat " " lem-ui-horiz-bar "\n "
+     (concat " " fedi-horiz-bar "\n "
              (upcase name)
-             "\n " lem-ui-horiz-bar "\n")
+             "\n " fedi-horiz-bar "\n")
      'face 'success)))
 
 (defun fedi-insert-heading (name)
@@ -282,7 +343,9 @@ text, i.e. hidden spoiler text."
     (replied   . ("⬇" . "↓"))
     (community . ("👪" . "[community]"))
     (reply-bar . ("│" . "|")) ;┃
-    (deleted   . ("🗑" . "[deleted]")))
+    (deleted   . ("🗑" . "[deleted]"))
+    (plus      . ("＋" . "+"))
+    (minus     . ("－" . "-")))
   "A set of symbols (and fallback strings) to be used in timeline.
 If a symbol does not look right (tofu), it means your
 font settings do not support it."
@@ -302,7 +365,7 @@ NAME is not part of the symbol table, '?' is returned."
 (defun fedi-font-lock-comment (&rest strs)
   "Font lock comment face STRS."
   (propertize (mapconcat #'identity strs "")
-              'face font-lock-comment-face))
+              'face 'font-lock-comment-face))
 
 (defun fedi-thing-json ()
   "Get json of thing at point, comment, post, community or user."
@@ -488,7 +551,7 @@ from the start if it is nil."
                                #'fedi--update-timestamps-callback
                                buffer nil))))))))
 
-(defun fedi--relative-time-details (timestamp &optional current-time)
+(defun fedi--relative-time-details (timestamp &optional current-time brief)
   "Return cons of (descriptive string . next change) for the TIMESTAMP.
 Use the optional CURRENT-TIME as the current time (only used for
 reliable testing).
@@ -510,10 +573,10 @@ TIMESTAMP is assumed to be in the past."
             (cons "just now"
                   60))
            ((< seconds-difference (* 1.5 60))
-            (cons "1 minute ago"
+            (cons (format "1 %s ago" (if brief "min" "minute"))
                   90)) ;; at 90 secs
            ((< seconds-difference (* 60 59.5))
-            (funcall regular-response seconds-difference 60 "minute"))
+            (funcall regular-response seconds-difference 60 (if brief "min" "minute")))
            ((< seconds-difference (* 1.5 60 60))
             (cons "1 hour ago"
                   (* 60 90))) ;; at 90 minutes
@@ -540,13 +603,13 @@ TIMESTAMP is assumed to be in the past."
     (cons (car relative-result)
           (time-add timestamp (seconds-to-time (cdr relative-result))))))
 
-(defun fedi--relative-time-description (timestamp &optional current-time)
+(defun fedi--relative-time-description (timestamp &optional current-time brief)
   "Return a string with a human readable TIMESTAMP relative to the current time.
 Use the optional CURRENT-TIME as the current time (only used for
 reliable testing).
 E.g. this could return something like \"1 min ago\", \"yesterday\", etc.
 TIME-STAMP is assumed to be in the past."
-  (car (fedi--relative-time-details timestamp current-time)))
+  (car (fedi--relative-time-details timestamp current-time brief)))
 
 ;;; LIVE BUFFERS
 
@@ -588,18 +651,24 @@ candidate's car, a string, usually its name or a handle."
          (completion-extra-properties
           (when list
             (list :annotation-function
-                  (lambda (i)
-                    (let ((annot (nth 2 (assoc i list #'equal))))
-                      (concat
-                       (propertize " " 'display
-                                   '(space :align-to (- right-margin 51)))
-                       (string-limit (car (string-lines annot)) 50)))))))
+                  (lambda (cand)
+                    (funcall #'fedi-annot-fun cand list)))))
          (choice (when list (completing-read prompt list)))
          (id (when list (nth 1 (assoc choice list #'equal)))))
     (if (not list)
         (user-error "No items returned")
       (funcall action-fun id choice))))
 
+(defun fedi-annot-fun (cand list)
+  "Annotation function for `fedi-do-item-completing'.
+Given CAND, return from LIST its annotation."
+  (let ((annot (nth 2
+                    (assoc cand list #'equal))))
+    (concat
+     (propertize " " 'display
+                 '(space :align-to (- right-margin 51)))
+     (string-limit (car (string-lines annot)) 50))))
+
 (defun fedi-response-msg (response &optional key value format-str)
   "Check RESPONSE, JSON from the server, and message on success.
 Used to handle server responses after the user
@@ -667,6 +736,7 @@ should be called with at least 1 arg: the item's json."
 We roll our own `replace-region-contents' because it is as
 non-destructive as possible, whereas we need to always replace
 the whole likes count in order to propertize it fully."
+  ;; can replace this w data arg, as we also get it in `fedi-update-item-from-json'?
   (let ((json (fedi--property 'json)))
     (save-excursion
       (goto-char beg)
@@ -674,5 +744,68 @@ the whole likes count in order to propertize it fully."
       (insert
        (funcall replace-fun json)))))
 
+;;; PROPERTIZING SPECIAL ITEMS
+
+(defun fedi-propertize-items (regex type keymap subexp
+                                    &optional item-subexp domain-subexp link
+                                    extra-props face)
+  "Propertize items of TYPE matching REGEX in STR as links using JSON.
+KEYMAP and LINK are properties to add to the match.
+EXTRA-PROPS is a property list of any extra properties to add.
+
+SUBEXP, ITEM-SUBEXP and DOMAIN-SUBEXP are regex subexpressions to
+handle submatches. Domain can be used to construct a URL, while
+the item submatch contains the item without a preceding @ or
+similar. It is added to the object's item property, so it can be
+used in a link function. For an example of regexes' subgroups, see
+`fedi-post-handle-regex'."
+  ;; FIXME: ideally we'd not do this in a sep buffer (gc)
+  ;; this runs on every item for every regex type!
+  (save-excursion
+    (save-match-data
+      (goto-char (point-min))
+      ;; ideally we'd work errors out, but we don't want to ruin
+      ;; our caller, which might make a page load fail:
+      (ignore-errors
+        ;; FIXME: do URLs
+        ;; (if (eq type 'url)
+        ;; (lem-ui-tabstop-link-by-regex regex)
+        (while (re-search-forward regex)
+          (let* ((item (when item-subexp
+                         (buffer-substring-no-properties
+                          (match-beginning item-subexp)
+                          (match-end item-subexp))))
+                 (beg (match-beginning subexp))
+                 (end (match-end subexp))
+                 (item-str (buffer-substring-no-properties beg end))
+                 (domain (when domain-subexp ; fedi-post-handle-regex
+                           (buffer-substring-no-properties (match-beginning domain-subexp)
+                                                           (match-end domain-subexp))))
+                 (link (cond ((eq type 'shr)
+                              item-str) ;; IF TYPE SHR: USE MATCHED URL
+                             ((functionp link)
+                              (funcall link))
+                             (t link))))
+            (add-text-properties beg
+                                 end
+                                 (fedi-link-props face link item type item-str keymap))
+            (add-text-properties beg end
+                                 extra-props)))))))
+
+(defun fedi-link-props (&optional face link item type help-echo keymap)
+  "Return a plist for a link.
+FACE LINK ITEM TYPE HELP-ECHO KEYMAP."
+  `(face ,(or face '(shr-text shr-link))
+         mouse-face highlight
+         shr-tab-stop t
+         shr-url ,link
+         button t
+         type ,type
+         item ,item
+         category shr
+         follow-link t
+         help-echo ,help-echo
+         keymap ,keymap))
+
 (provide 'fedi)
 ;;; fedi.el ends here
